@@ -68,18 +68,24 @@
   "Show window ID exactly fit in the Emacs window WINDOW."
   (exwm--log "Show #x%x in %s" id window)
   (let* ((edges (window-inside-absolute-pixel-edges window))
-         (width (- (elt edges 2) (elt edges 0)))
-         (height (- (elt edges 3) (elt edges 1)))
+         (x (pop edges))
+         (y (pop edges))
+         (width (- (pop edges) x))
+         (height (- (pop edges) y))
+         (edges (window-inside-pixel-edges window))
+         (relative-x (pop edges))
+         (relative-y (pop edges))
          frame-width frame-height)
     (with-current-buffer (exwm--id->buffer id)
       (if (not exwm--floating-frame)
-          (let ((relative-edges (window-inside-pixel-edges window)))
-            (exwm-layout--resize-container
-             id exwm--container
-             (elt relative-edges 0) (elt relative-edges 1) width height
-             ;; Do not resize the X window if the minibuffer resizes itself.
-             (and (active-minibuffer-window)
-                  (< 1 (window-height (active-minibuffer-window))))))
+          (exwm-layout--resize-container id exwm--container
+                                         relative-x relative-y width height
+                                         ;; Keep the size of the X window if
+                                         ;; it's the minibuffer that resized.
+                                         (and
+                                          (active-minibuffer-window)
+                                          (< 1 (window-height
+                                                (active-minibuffer-window)))))
         ;; A floating X window is of the same size as the Emacs window,
         ;; whereas its container is of the same size as the Emacs frame.
         (setq frame-width (frame-pixel-width exwm--floating-frame)
@@ -106,8 +112,8 @@
                                                xcb:ConfigWindow:Y
                                                xcb:ConfigWindow:Width
                                                xcb:ConfigWindow:Height)
-                           :x exwm-floating-border-width
-                           :y exwm-floating-border-width
+                           :x relative-x
+                           :y relative-y
                            :width width
                            :height height)))
       ;; Make the resizing take effect.
@@ -122,11 +128,13 @@
                        :event-mask xcb:EventMask:StructureNotify
                        :event (xcb:marshal
                                (make-instance 'xcb:ConfigureNotify
-                                              :event id :window id
+                                              :event id
+                                              :window id
                                               :above-sibling xcb:Window:None
-                                              :x (elt edges 0)
-                                              :y (elt edges 1)
-                                              :width width :height height
+                                              :x x
+                                              :y y
+                                              :width width
+                                              :height height
                                               :border-width 0
                                               :override-redirect 0)
                                exwm--connection))))
@@ -169,7 +177,7 @@
   "Make window ID fullscreen."
   (interactive)
   (with-current-buffer (if id (exwm--id->buffer id) (window-buffer))
-    (when exwm--fullscreen
+    (when (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state)
       (user-error "Already in full-screen mode."))
     ;; Save the position of floating frame.
     (when exwm--floating-frame
@@ -213,7 +221,7 @@
                        :window exwm--id
                        :data (vector xcb:Atom:_NET_WM_STATE_FULLSCREEN)))
     (xcb:flush exwm--connection)
-    (setq exwm--fullscreen t)
+    (cl-pushnew xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state)
     (call-interactively #'exwm-input-release-keyboard)))
 
 ;;;###autoload
@@ -221,7 +229,7 @@
   "Restore window from fullscreen state."
   (interactive)
   (with-current-buffer (if id (exwm--id->buffer id) (window-buffer))
-    (unless exwm--fullscreen
+    (unless (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state)
       (user-error "Not in full-screen mode."))
     ;; Restore the size of this workspace.
     (exwm-workspace--set-fullscreen exwm--frame)
@@ -248,7 +256,8 @@
     (xcb:+request exwm--connection
         (make-instance 'xcb:ewmh:set-_NET_WM_STATE :window exwm--id :data []))
     (xcb:flush exwm--connection)
-    (setq exwm--fullscreen nil)
+    (setq exwm--ewmh-state
+          (delq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state))
     (call-interactively #'exwm-input-grab-keyboard)))
 
 (defvar exwm-layout--other-buffer-exclude-exwm-mode-buffers nil
@@ -306,11 +315,14 @@ selected by `other-buffer'."
                        :data (vconcat (append clients-other clients-iconic
                                               clients clients-floating))))))
 
-(defun exwm-layout--refresh ()
+(defun exwm-layout--refresh (&optional frame)
   "Refresh layout."
-  (let ((frame (selected-frame))
-        covered-buffers             ;EXWM-buffers covered by a new X window.
-        vacated-windows             ;Windows previously displaying EXWM-buffers.
+  ;; `window-size-change-functions' sets this argument while
+  ;; `window-configuration-change-hook' makes the frame selected.
+  (unless frame
+    (setq frame (selected-frame)))
+  (let (covered-buffers   ;EXWM-buffers covered by a new X window.
+        vacated-windows   ;Windows previously displaying EXWM-buffers.
         windows)
     (if (not (exwm-workspace--workspace-p frame))
         (if (frame-parameter frame 'exwm-outer-id)
@@ -375,9 +387,12 @@ selected by `other-buffer'."
       (exwm-layout--set-client-list-stacking)
       (xcb:flush exwm--connection))))
 
+(declare-function exwm-workspace--client-p "exwm-workspace.el"
+                  (&optional frame))
+
 (defun exwm-layout--on-minibuffer-setup ()
   "Refresh layout when minibuffer grows."
-  (when (frame-parameter nil 'exwm-outer-id)
+  (unless (exwm-workspace--client-p)
     (run-with-idle-timer 0.01 nil         ;FIXME
                          (lambda ()
                            (when (< 1 (window-height (minibuffer-window)))
@@ -386,8 +401,7 @@ selected by `other-buffer'."
 (defun exwm-layout--on-echo-area-change (&optional dirty)
   "Run when message arrives or in `echo-area-clear-hook' to refresh layout."
   (when (and (current-message)
-             ;; Exclude non-graphical frames.
-             (frame-parameter nil 'exwm-outer-id)
+             (not (exwm-workspace--client-p))
              (or (cl-position ?\n (current-message))
                  (> (length (current-message))
                     (frame-width exwm-workspace--current))))
@@ -556,6 +570,9 @@ See also `exwm-layout-enlarge-window'."
   "Initialize layout module."
   ;; Auto refresh layout
   (add-hook 'window-configuration-change-hook #'exwm-layout--refresh)
+  ;; The behavior of `window-configuration-change-hook' will be changed.
+  (when (fboundp 'window-pixel-width-before-size-change)
+    (add-hook 'window-size-change-functions #'exwm-layout--refresh))
   (unless (exwm-workspace--minibuffer-own-frame-p)
     ;; Refresh when minibuffer grows
     (add-hook 'minibuffer-setup-hook #'exwm-layout--on-minibuffer-setup t)
