@@ -25,8 +25,12 @@
 
 ;;; Code:
 
+(require 'server)
+
 (require 'exwm-core)
 
+(defvar exwm-manage--desktop)
+
 (defvar exwm-workspace-number 1 "Initial number of workspaces.")
 (defvar exwm-workspace--list nil "List of all workspaces (Emacs frames).")
 (defvar exwm-workspace--current nil "Current active workspace.")
@@ -46,6 +50,10 @@ NIL if FRAME is not a workspace"
   "Return t if FRAME is a workspace."
   (memq frame exwm-workspace--list))
 
+(defsubst exwm-workspace--client-p (&optional frame)
+  "Return non-nil if FRAME is an emacsclient frame."
+  (frame-parameter frame 'client))
+
 (defun exwm-workspace--workspace-from-frame-or-index (frame-or-index)
   "Retrieve the workspace frame from FRAME-OR-INDEX."
   (cond
@@ -224,18 +232,19 @@ Value nil means to use the default position which is fixed at bottom, while
   (let (struts struts*)
     (dolist (pair exwm-workspace--id-struts-alist)
       (setq struts (cdr pair))
-      (dotimes (i 4)
-        (when (/= 0 (aref struts i))
-          (setq struts*
-                (vector (aref [left right top bottom] i)
-                        (aref struts i)
-                        (when (= 12 (length struts))
-                          (substring struts (+ 4 (* i 2)) (+ 6 (* i 2))))))
-          (if (= 0 (mod i 2))
-              ;; Make left/top processed first.
-              (push struts* exwm-workspace--struts)
-            (setq exwm-workspace--struts
-                  (append exwm-workspace--struts (list struts*)))))))))
+      (when struts
+        (dotimes (i 4)
+          (when (/= 0 (aref struts i))
+            (setq struts*
+                  (vector (aref [left right top bottom] i)
+                          (aref struts i)
+                          (when (= 12 (length struts))
+                            (substring struts (+ 4 (* i 2)) (+ 6 (* i 2))))))
+            (if (= 0 (mod i 2))
+                ;; Make left/top processed first.
+                (push struts* exwm-workspace--struts)
+              (setq exwm-workspace--struts
+                    (append exwm-workspace--struts (list struts*))))))))))
 
 (defvar exwm-workspace--workareas nil "Workareas (struts excluded).")
 
@@ -363,11 +372,17 @@ If the minibuffer is detached, this value is 0.")
                        :value-mask (logior xcb:ConfigWindow:X
                                            xcb:ConfigWindow:Y
                                            xcb:ConfigWindow:Width
+                                           (if exwm-manage--desktop
+                                               xcb:ConfigWindow:Sibling
+                                             0)
                                            xcb:ConfigWindow:StackMode)
                        :x (aref workarea 0)
                        :y y
                        :width width
-                       :stack-mode xcb:StackMode:Below))
+                       :sibling exwm-manage--desktop
+                       :stack-mode (if exwm-manage--desktop
+                                       xcb:StackMode:Above
+                                     xcb:StackMode:Below)))
     (xcb:+request exwm--connection
         (make-instance 'xcb:ConfigureWindow
                        :window (frame-parameter exwm-workspace--minibuffer
@@ -431,7 +446,9 @@ PREFIX-DIGITS is a list of the digits introduced so far."
 The optional FORCE option is for internal use only."
   (interactive
    (list
-    (unless (and (eq major-mode 'exwm-mode) exwm--fullscreen) ;it's invisible
+    (unless (and (eq major-mode 'exwm-mode)
+                 ;; The prompt is invisible in fullscreen mode.
+                 (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state))
       (let ((exwm-workspace--prompt-add-allowed t)
             (exwm-workspace--prompt-delete-allowed t))
         (exwm-workspace--prompt-for-workspace "Switch to [+/-]: ")))))
@@ -449,7 +466,8 @@ The optional FORCE option is for internal use only."
                          :value-mask xcb:ConfigWindow:StackMode
                          :stack-mode xcb:StackMode:Above))
       ;; Raise X windows with struts set if there's no fullscreen X window.
-      (unless (buffer-local-value 'exwm--fullscreen (window-buffer window))
+      (unless (with-current-buffer (window-buffer window)
+                (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state))
         (dolist (pair exwm-workspace--id-struts-alist)
           (xcb:+request exwm--connection
               (make-instance 'xcb:ConfigureWindow
@@ -523,7 +541,9 @@ deleted, moved, etc).")
 (defun exwm-workspace-swap (workspace1 workspace2)
   "Interchange position of WORKSPACE1 with that of WORKSPACE2."
   (interactive
-   (unless (and (eq major-mode 'exwm-mode) exwm--fullscreen) ;it's invisible
+   (unless (and (eq major-mode 'exwm-mode)
+                ;; The prompt is invisible in fullscreen mode.
+                (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state))
      (let (w1 w2)
        (let ((exwm-workspace--prompt-add-allowed t)
              (exwm-workspace--prompt-delete-allowed t))
@@ -558,7 +578,9 @@ deleted, moved, etc).")
 When called interactively, prompt for a workspace and move current one just
 before it."
   (interactive
-   (unless (and (eq major-mode 'exwm-mode) exwm--fullscreen) ;it's invisible
+   (unless (and (eq major-mode 'exwm-mode)
+                ;; The prompt is invisible in fullscreen mode.
+                (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state))
      (list exwm-workspace--current
            (exwm-workspace--position
             (exwm-workspace--prompt-for-workspace "Move workspace to: ")))))
@@ -685,10 +707,13 @@ INDEX must not exceed the current number of workspaces."
                        (outer-id (string-to-number
                                   (frame-parameter new-frame
                                                    'outer-window-id)))
+                       (window-id (string-to-number
+                                   (frame-parameter new-frame 'window-id)))
                        (frame-container (frame-parameter old-frame
                                                          'exwm-container))
                        (window (frame-root-window new-frame)))
                   (set-frame-parameter new-frame 'exwm-outer-id outer-id)
+                  (set-frame-parameter new-frame 'exwm-id window-id)
                   (set-frame-parameter new-frame 'exwm-container
                                        frame-container)
                   (make-frame-invisible new-frame)
@@ -880,24 +905,25 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
 
 (defun exwm-workspace--update-minibuffer-height (&optional echo-area)
   "Update the minibuffer frame height."
-  (let ((height
-         (with-current-buffer
-             (window-buffer (minibuffer-window exwm-workspace--minibuffer))
-           (max 1
-                (if echo-area
-                    (let ((width (frame-width exwm-workspace--minibuffer))
-                          (result 0))
-                      (mapc (lambda (i)
-                              (setq result
-                                    (+ result
-                                       (ceiling (1+ (length i)) width))))
-                            (split-string (or (current-message) "") "\n"))
-                      result)
-                  (count-screen-lines))))))
-    (when (and (integerp max-mini-window-height)
-               (> height max-mini-window-height))
-      (setq height max-mini-window-height))
-    (set-frame-height exwm-workspace--minibuffer height)))
+  (unless (exwm-workspace--client-p)
+    (let ((height
+           (with-current-buffer
+               (window-buffer (minibuffer-window exwm-workspace--minibuffer))
+             (max 1
+                  (if echo-area
+                      (let ((width (frame-width exwm-workspace--minibuffer))
+                            (result 0))
+                        (mapc (lambda (i)
+                                (setq result
+                                      (+ result
+                                         (ceiling (1+ (length i)) width))))
+                              (split-string (or (current-message) "") "\n"))
+                        result)
+                    (count-screen-lines))))))
+      (when (and (integerp max-mini-window-height)
+                 (> height max-mini-window-height))
+        (setq height max-mini-window-height))
+      (set-frame-height exwm-workspace--minibuffer height))))
 
 (defun exwm-workspace--on-ConfigureNotify (data _synthetic)
   "Adjust the container to fit the minibuffer frame."
@@ -977,15 +1003,20 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
       (make-instance 'xcb:ConfigureWindow
                      :window (frame-parameter exwm-workspace--minibuffer
                                               'exwm-container)
-                     :value-mask xcb:ConfigWindow:StackMode
-                     :stack-mode xcb:StackMode:Below))
+                     :value-mask (logior (if exwm-manage--desktop
+                                             xcb:ConfigWindow:Sibling
+                                           0)
+                                         xcb:ConfigWindow:StackMode)
+                     :sibling exwm-manage--desktop
+                     :stack-mode (if exwm-manage--desktop
+                                     xcb:StackMode:Above
+                                   xcb:StackMode:Below)))
   (xcb:flush exwm--connection))
 
 (defun exwm-workspace--on-minibuffer-setup ()
   "Run in minibuffer-setup-hook to show the minibuffer and its container."
   (when (and (= 1 (minibuffer-depth))
-             ;; Exclude non-graphical frames.
-             (frame-parameter nil 'exwm-outer-id))
+             (not (exwm-workspace--client-p)))
     (add-hook 'post-command-hook #'exwm-workspace--update-minibuffer-height)
     (exwm-workspace--show-minibuffer))
   ;; FIXME: This is a temporary fix for the *Completions* buffer not
@@ -1006,8 +1037,7 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
 (defun exwm-workspace--on-minibuffer-exit ()
   "Run in minibuffer-exit-hook to hide the minibuffer container."
   (when (and (= 1 (minibuffer-depth))
-             ;; Exclude non-graphical frames.
-             (frame-parameter nil 'exwm-outer-id))
+             (not (exwm-workspace--client-p)))
     (remove-hook 'post-command-hook #'exwm-workspace--update-minibuffer-height)
     (exwm-workspace--hide-minibuffer)))
 
@@ -1016,8 +1046,7 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
 (defun exwm-workspace--on-echo-area-dirty ()
   "Run when new message arrives to show the echo area and its container."
   (when (and (not (active-minibuffer-window))
-             ;; Exclude non-graphical frames.
-             (frame-parameter nil 'exwm-outer-id)
+             (not (exwm-workspace--client-p))
              (or (current-message)
                  cursor-in-echo-area))
     (exwm-workspace--update-minibuffer-height t)
@@ -1031,7 +1060,7 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
 
 (defun exwm-workspace--on-echo-area-clear ()
   "Run in echo-area-clear-hook to hide echo area container."
-  (when (frame-parameter nil 'exwm-outer-id) ;Exclude non-graphical frames.
+  (unless (exwm-workspace--client-p)
     (unless (active-minibuffer-window)
       (exwm-workspace--hide-minibuffer))
     (when exwm-workspace--display-echo-area-timer
@@ -1046,15 +1075,32 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
 
 (defun exwm-workspace--confirm-kill-emacs (prompt &optional force)
   "Confirm before exiting Emacs."
-  (when (or (and force (not (eq force 'no-check)))
-            (and (or (eq force 'no-check) (not exwm--id-buffer-alist))
-                 (y-or-n-p prompt))
-            (yes-or-no-p (format "[EXWM] %d window(s) will be destroyed.  %s"
-                                 (length exwm--id-buffer-alist) prompt)))
-    ;; Run `kill-emacs-hook' before Emacs frames are unmapped so that
-    ;; errors can be visible.
-    (run-hooks 'kill-emacs-hook)
-    (setq kill-emacs-hook nil)
+  (when (cond
+         ((and force (not (eq force 'no-check)))
+          ;; Force killing Emacs.
+          t)
+         ((or (eq force 'no-check) (not exwm--id-buffer-alist))
+          ;; Check if there's any unsaved file.
+          (pcase (catch 'break
+                   (let ((kill-emacs-query-functions
+                          (append kill-emacs-query-functions
+                                  (list (lambda ()
+                                          (throw 'break 'break))))))
+                     (save-buffers-kill-emacs)))
+            (`break (y-or-n-p prompt))
+            (x x)))
+         (t
+          (yes-or-no-p (format "[EXWM] %d window(s) will be destroyed.  %s"
+                               (length exwm--id-buffer-alist) prompt))))
+    ;; Run `kill-emacs-hook' (`server-force-stop' excluded) before Emacs
+    ;; frames are unmapped so that errors (if any) can be visible.
+    (if (memq #'server-force-stop kill-emacs-hook)
+        (progn
+          (setq kill-emacs-hook (delq #'server-force-stop kill-emacs-hook))
+          (run-hooks 'kill-emacs-hook)
+          (setq kill-emacs-hook (list #'server-force-stop)))
+      (run-hooks 'kill-emacs-hook)
+      (setq kill-emacs-hook nil))
     ;; Hide & reparent out all frames (save-set can't be used here since
     ;; X windows will be re-mapped).
     (when (exwm-workspace--minibuffer-own-frame-p)
@@ -1079,22 +1125,20 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
                            :parent exwm--root
                            :x 0
                            :y 0))))
-    ;; Exit each module.
-    (exwm--exit)
-    ;; Destroy all resources created by this connection.
-    (xcb:disconnect exwm--connection)
-    (setq exwm--connection nil)
-    ;; Extra cleanups for emacsclient.
+    ;; Restore the 'client' frame parameter (before `exwm--exit').
     (when exwm-workspace--client
       (dolist (f exwm-workspace--list)
         (set-frame-parameter f 'client exwm-workspace--client))
       (when (exwm-workspace--minibuffer-own-frame-p)
         (set-frame-parameter exwm-workspace--minibuffer 'client
-                             exwm-workspace--client))
-      ;; Kill the client.
-      (server-save-buffers-kill-terminal nil))
+                             exwm-workspace--client)))
+    ;; Exit each module.
+    (exwm--exit)
+    ;; Destroy all resources created by this connection.
+    (xcb:disconnect exwm--connection)
+    (setq exwm--connection nil)
     ;; Set the return value.
-    (not exwm-workspace--client)))
+    t))
 
 (defun exwm-workspace--set-desktop-geometry ()
   "Set _NET_DESKTOP_GEOMETRY."
@@ -1130,12 +1174,16 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
     (setq exwm-workspace--list (nconc exwm-workspace--list (list frame)))
     (let ((outer-id (string-to-number (frame-parameter frame
                                                        'outer-window-id)))
+          (window-id (string-to-number (frame-parameter frame 'window-id)))
           (container (xcb:generate-id exwm--connection))
           (workspace (xcb:generate-id exwm--connection)))
       ;; Save window IDs
       (set-frame-parameter frame 'exwm-outer-id outer-id)
+      (set-frame-parameter frame 'exwm-id window-id)
       (set-frame-parameter frame 'exwm-container container)
       (set-frame-parameter frame 'exwm-workspace workspace)
+      ;; In case it's created by emacsclient.
+      (set-frame-parameter frame 'client nil)
       ;; Copy RandR frame parameters from the first workspace to
       ;; prevent potential problems.  The values do not matter here as
       ;; they'll be updated by the RandR module later.
@@ -1145,14 +1193,20 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
           (set-frame-parameter frame param (frame-parameter w param))))
       (xcb:+request exwm--connection
           (make-instance 'xcb:CreateWindow
-                         :depth 0 :wid workspace :parent exwm--root
-                         :x 0 :y 0
+                         :depth 0
+                         :wid workspace
+                         :parent exwm--root
+                         :x 0
+                         :y 0
                          :width (x-display-pixel-width)
                          :height (x-display-pixel-height)
-                         :border-width 0 :class xcb:WindowClass:CopyFromParent
-                         :visual 0      ;CopyFromParent
-                         :value-mask (logior xcb:CW:OverrideRedirect
+                         :border-width 0
+                         :class xcb:WindowClass:InputOutput
+                         :visual 0
+                         :value-mask (logior xcb:CW:BackPixmap
+                                             xcb:CW:OverrideRedirect
                                              xcb:CW:EventMask)
+                         :background-pixmap xcb:BackPixmap:ParentRelative
                          :override-redirect 1
                          :event-mask xcb:EventMask:SubstructureRedirect))
       (xcb:+request exwm--connection
@@ -1162,13 +1216,19 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
                          :stack-mode xcb:StackMode:Below))
       (xcb:+request exwm--connection
           (make-instance 'xcb:CreateWindow
-                         :depth 0 :wid container :parent workspace
-                         :x 0 :y 0
+                         :depth 0
+                         :wid container
+                         :parent workspace
+                         :x 0
+                         :y 0
                          :width (x-display-pixel-width)
                          :height (x-display-pixel-height)
-                         :border-width 0 :class xcb:WindowClass:CopyFromParent
-                         :visual 0      ;CopyFromParent
-                         :value-mask xcb:CW:OverrideRedirect
+                         :border-width 0
+                         :class xcb:WindowClass:InputOutput
+                         :visual 0
+                         :value-mask (logior xcb:CW:BackPixmap
+                                             xcb:CW:OverrideRedirect)
+                         :background-pixmap xcb:BackPixmap:ParentRelative
                          :override-redirect 1))
       (exwm--debug
        (xcb:+request exwm--connection
@@ -1260,11 +1320,6 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
                        :window exwm--root :data num-workspaces))
     ;; Set _NET_DESKTOP_GEOMETRY.
     (exwm-workspace--set-desktop-geometry)
-    ;; Set _NET_DESKTOP_VIEWPORT (we don't support large desktop).
-    (xcb:+request exwm--connection
-        (make-instance 'xcb:ewmh:set-_NET_DESKTOP_VIEWPORT
-                       :window exwm--root
-                       :data (make-vector (* 2 num-workspaces) 0)))
     ;; Update and set _NET_WORKAREA.
     (exwm-workspace--update-workareas)
     ;; Set _NET_VIRTUAL_ROOTS.
@@ -1307,7 +1362,7 @@ applied to all subsequently created X frames."
             (unless (frame-parameter i 'window-id)
               (setq initial-workspaces (delq i initial-workspaces))))
           (setq exwm-workspace--client
-                (frame-parameter (car exwm-workspace--list) 'client))
+                (frame-parameter (car initial-workspaces) 'client))
           (let ((f (car initial-workspaces)))
             ;; Remove the possible internal border.
             (set-frame-parameter f 'internal-border-width 0)
@@ -1335,19 +1390,30 @@ applied to all subsequently created X frames."
       (let ((outer-id (string-to-number
                        (frame-parameter exwm-workspace--minibuffer
                                         'outer-window-id)))
+            (window-id (string-to-number
+                        (frame-parameter exwm-workspace--minibuffer
+                                         'window-id)))
             (container (xcb:generate-id exwm--connection)))
         (set-frame-parameter exwm-workspace--minibuffer
                              'exwm-outer-id outer-id)
+        (set-frame-parameter exwm-workspace--minibuffer 'exwm-id window-id)
         (set-frame-parameter exwm-workspace--minibuffer 'exwm-container
                              container)
         (xcb:+request exwm--connection
             (make-instance 'xcb:CreateWindow
-                           :depth 0 :wid container :parent exwm--root
-                           :x -1 :y -1 :width 1 :height 1
+                           :depth 0
+                           :wid container
+                           :parent exwm--root
+                           :x 0
+                           :y 0
+                           :width 1
+                           :height 1
                            :border-width 0
-                           :class xcb:WindowClass:CopyFromParent
-                           :visual 0        ;CopyFromParent
-                           :value-mask xcb:CW:OverrideRedirect
+                           :class xcb:WindowClass:InputOutput
+                           :visual 0
+                           :value-mask (logior xcb:CW:BackPixmap
+                                               xcb:CW:OverrideRedirect)
+                           :background-pixmap xcb:BackPixmap:ParentRelative
                            :override-redirect 1))
         (exwm--debug
          (xcb:+request exwm--connection
