@@ -4,8 +4,8 @@
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 ;; Maintainer: Chris Feng <chris.w.feng@gmail.com>
-;; Version: 0.8
-;; Package-Requires: ((xelb "0.10"))
+;; Version: 0.9
+;; Package-Requires: ((xelb "0.11"))
 ;; Keywords: unix
 ;; URL: https://github.com/ch11ng/exwm
 
@@ -78,7 +78,8 @@
   (interactive)
   (with-current-buffer (window-buffer)
     (when (eq major-mode 'exwm-mode)
-      (when exwm--fullscreen (exwm-layout-unset-fullscreen))
+      (when (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN exwm--ewmh-state)
+        (exwm-layout-unset-fullscreen))
       ;; Force refresh
       (exwm-layout--refresh)
       (call-interactively #'exwm-input-grab-keyboard))))
@@ -88,10 +89,29 @@
   "Restart EXWM."
   (interactive)
   (when (exwm-workspace--confirm-kill-emacs "[EXWM] Restart? " 'no-check)
-    (apply #'call-process (car command-line-args) nil nil nil
-           (cdr command-line-args))
-    ;; Kill this instance at last.
-    (kill-emacs)))
+    (let* ((attr (process-attributes (emacs-pid)))
+           (args (cdr (assq 'args attr)))
+           (ppid (cdr (assq 'ppid attr)))
+           (pargs (cdr (assq 'args (process-attributes ppid)))))
+      (cond
+       ((= ppid 1)
+        ;; The parent is the init process.  This probably means this
+        ;; instance is an emacsclient.  Anyway, start a control instance
+        ;; to manage the subsequent ones.
+        (call-process (car command-line-args))
+        (kill-emacs))
+       ((string= args pargs)
+        ;; This is a subordinate instance.  Return a magic number to
+        ;; inform the parent (control instance) to start another one.
+        (kill-emacs ?R))
+       (t
+        ;; This is the control instance.  Keep starting subordinate
+        ;; instances until told to exit.
+        ;; Run `server-force-stop' if it exists.
+        (run-hooks 'kill-emacs-hook)
+        (with-temp-buffer
+          (while (= ?R (shell-command-on-region (point) (point) args))))
+        (kill-emacs))))))
 
 (defun exwm--update-window-type (id &optional force)
   "Update _NET_WM_WINDOW_TYPE."
@@ -248,13 +268,9 @@
                                      :window id)))
       (when reply
         (setq struts (slot-value reply 'value))
-        (if struts
-            (if pair
-                (setcdr pair struts)
-              (push (cons id struts) exwm-workspace--id-struts-alist))
-          (when pair
-            (setq exwm-workspace--id-struts-alist
-                  (assq-delete-all id exwm-workspace--id-struts-alist))))
+        (if pair
+            (setcdr pair struts)
+          (push (cons id struts) exwm-workspace--id-struts-alist))
         (exwm-workspace--update-struts))
       ;; Update workareas and set _NET_WORKAREA.
       (exwm-workspace--update-workareas)
@@ -271,13 +287,9 @@
     (when reply
       (setq struts (slot-value reply 'value)
             pair (assq id exwm-workspace--id-struts-alist))
-      (if struts
-          (if pair
-              (setcdr pair struts)
-            (push (cons id struts) exwm-workspace--id-struts-alist))
-        (when pair
-          (setq exwm-workspace--id-struts-alist
-                (assq-delete-all id exwm-workspace--id-struts-alist))))
+      (if pair
+          (setcdr pair struts)
+        (push (cons id struts) exwm-workspace--id-struts-alist))
       (exwm-workspace--update-struts))
     ;; Update workareas and set _NET_WORKAREA.
     (exwm-workspace--update-workareas)
@@ -296,8 +308,7 @@
         atom id buffer)
     (xcb:unmarshal obj data)
     (setq id (slot-value obj 'window)
-          atom (slot-value obj 'atom)
-          exwm-input--timestamp (slot-value obj 'time))
+          atom (slot-value obj 'atom))
     (setq buffer (exwm--id->buffer id))
     (if (not (buffer-live-p buffer))
         ;; Properties of unmanaged X windows.
@@ -446,12 +457,17 @@
             (when (or (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN props)
                       (memq xcb:Atom:_NET_WM_STATE_ABOVE props))
               (cond ((= action xcb:ewmh:_NET_WM_STATE_ADD)
-                     (unless exwm--fullscreen (exwm-layout-set-fullscreen id))
+                     (unless (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN
+                                   exwm--ewmh-state)
+                       (exwm-layout-set-fullscreen id))
                      (push xcb:Atom:_NET_WM_STATE_FULLSCREEN props-new))
                     ((= action xcb:ewmh:_NET_WM_STATE_REMOVE)
-                     (when exwm--fullscreen (exwm-layout-unset-fullscreen id)))
+                     (when (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN
+                                 exwm--ewmh-state)
+                       (exwm-layout-unset-fullscreen id)))
                     ((= action xcb:ewmh:_NET_WM_STATE_TOGGLE)
-                     (if exwm--fullscreen
+                     (if (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN
+                               exwm--ewmh-state)
                          (exwm-layout-unset-fullscreen id)
                        (exwm-layout-set-fullscreen id)
                        (push xcb:Atom:_NET_WM_STATE_FULLSCREEN props-new)))))
@@ -591,10 +607,17 @@
   (let ((new-id (xcb:generate-id exwm--connection)))
     (xcb:+request exwm--connection
         (make-instance 'xcb:CreateWindow
-                       :depth 0 :wid new-id :parent exwm--root
-                       :x -1 :y -1 :width 1 :height 1
-                       :border-width 0 :class xcb:WindowClass:CopyFromParent
-                       :visual 0 :value-mask xcb:CW:OverrideRedirect
+                       :depth 0
+                       :wid new-id
+                       :parent exwm--root
+                       :x 0
+                       :y 0
+                       :width 1
+                       :height 1
+                       :border-width 0
+                       :class xcb:WindowClass:InputOnly
+                       :visual 0
+                       :value-mask xcb:CW:OverrideRedirect
                        :override-redirect 1))
     (dolist (i (list exwm--root new-id))
       ;; Set _NET_SUPPORTING_WM_CHECK
@@ -605,6 +628,11 @@
       (xcb:+request exwm--connection
           (make-instance 'xcb:ewmh:set-_NET_WM_NAME
                          :window i :data "EXWM"))))
+  ;; Set _NET_DESKTOP_VIEWPORT (we don't support large desktop).
+  (xcb:+request exwm--connection
+      (make-instance 'xcb:ewmh:set-_NET_DESKTOP_VIEWPORT
+                     :window exwm--root
+                     :data [0 0]))
   (xcb:flush exwm--connection))
 
 (defun exwm--exit-icccm-ewmh ()
